/* 
Demonstration of using the XAnimator library
By Keith Ditchburn 2010
*/

#define VC_EXTRALEAN
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

// XAnimator header
#include <XAnimator_lib.h>


#if defined(DEBUG) | defined(_DEBUG)
	#include <crtdbg.h>
	#define new new(_NORMAL_BLOCK,__FILE__, __LINE__)
#endif

// Direct3D objects
LPDIRECT3D9 gD3dObject=NULL;
LPDIRECT3DDEVICE9 gD3dDevice=NULL;

// Globals
IXAnimator *gXAnimator=0;
int gModelId=0;
int gNumAnimationSets=0;
int gCurrentAnimationSet=0;

// Function prototypes
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
bool SetupDirect3D(HWND hWnd);
void ShutdownDirect3D();
void UpdateDemo();
bool LoadModel(const std::string &filename);

// Constants that can be tweaked:

// Transition time between animation sets in ms
float kTransitionTime=2500;

// Time before we switch to the next animation in ms
DWORD kTimeBeforeNextAnimation=4000;

// Windows API program entry point
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
		// Leak checking 
#if defined(DEBUG) | defined(_DEBUG)
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc= (WNDPROC)WndProc;
	wcex.cbClsExtra= 0;
	wcex.cbWndExtra= 0;
	wcex.hInstance= hInstance;
	wcex.hIcon= 0;
	wcex.hCursor= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName= 0;
	wcex.lpszClassName= "XAnimatorTestWindowClass";
	wcex.hIconSm = 0;

	RegisterClassEx(&wcex);

	HWND hWnd = CreateWindow("XAnimatorTestWindowClass", "XAnimatorTest, Keith Ditchburn 2010", WS_OVERLAPPEDWINDOW, 
		CW_USEDEFAULT, CW_USEDEFAULT, 800, 600, 
		NULL, NULL, hInstance, NULL);

	// Create the D3D objects etc.
	if (!SetupDirect3D(hWnd))
		return 0;

	// Init XAnimator
	gXAnimator=CreateXAnimator(gD3dDevice);
	if (!gXAnimator)
		return 0;

	// Load our animated model
	if (!LoadModel("Data\\bones_all.x"))
		return 0;

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	// Message loop that calls our update function whenever there is no message
	MSG msg;
	ZeroMemory( &msg, sizeof(msg) );
	while( msg.message!=WM_QUIT )
	{
		if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
		else
		{
			UpdateDemo();
		}
	}

	// Clean up
	delete gXAnimator;
	ShutdownDirect3D();
	
	return (int)msg.wParam;
}

// Window callback
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) 
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}

	return 0;
}

// Setup Direct3D, matrices and add a light
bool SetupDirect3D(HWND hWnd)
{
	gD3dObject=Direct3DCreate9(D3D_SDK_VERSION);
	if (!gD3dObject)
		return false;

	D3DPRESENT_PARAMETERS presParams;
	ZeroMemory(&presParams,sizeof(presParams));
	presParams.Windowed=TRUE;
	presParams.SwapEffect=D3DSWAPEFFECT_DISCARD;
	presParams.BackBufferFormat=D3DFMT_UNKNOWN;
	presParams.PresentationInterval=D3DPRESENT_INTERVAL_ONE;
	presParams.EnableAutoDepthStencil = TRUE;
	presParams.AutoDepthStencilFormat = D3DFMT_D16;

	// Try to get the best device possible e.g. harwdare processing
	HRESULT hr=gD3dObject->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,hWnd, 
		D3DCREATE_HARDWARE_VERTEXPROCESSING, &presParams, &gD3dDevice);
	if (FAILED(hr))
	{
		// the above failed so fallback to software processing
		hr=gD3dObject->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,hWnd,
			D3DCREATE_SOFTWARE_VERTEXPROCESSING, &presParams, &gD3dDevice);
		if (FAILED(hr))
			return false;
	}

	gD3dDevice->SetRenderState( D3DRS_ZENABLE, D3DZB_TRUE );

	D3DXMATRIX matView;
	D3DXMatrixLookAtLH(&matView,&D3DXVECTOR3(0,1,-4),&D3DXVECTOR3(0,1,0), &D3DXVECTOR3(0,1,0) );
	gD3dDevice->SetTransform( D3DTS_VIEW, &matView );

	D3DXMATRIX matProj;
	D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, 800.0f/600.0f, 1.0f, 200.0f );
	gD3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

	gD3dDevice->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_XRGB(80,80,80) );

	// Fill in a light structure defining our light
	D3DLIGHT9 light;
	ZeroMemory( &light, sizeof(D3DLIGHT9) );
	light.Type       = D3DLIGHT_DIRECTIONAL;
	light.Diffuse.r  = 1.0f;
	light.Diffuse.g  = 1.0f;
	light.Diffuse.b  = 1.0f;
	light.Diffuse.a  = 1.0f;
	light.Range      = 500.0f;
	// Create a direction for our light - it must be normalized  
	D3DXVECTOR3 vecDir;
	vecDir = D3DXVECTOR3(0.0f,-0.3f,0.5);
	D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );

	// Tell the device about the light and turn it on
	gD3dDevice->SetLight( 0, &light );
	gD3dDevice->LightEnable( 0, TRUE ); 

	gD3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE );

	return true;
}

// Release the D3D objects
void ShutdownDirect3D()
{
	if (gD3dDevice)
	{
		gD3dDevice->Release();
		gD3dDevice=0;
	}
	if (gD3dObject)
	{
		gD3dObject->Release();
		gD3dObject=0;
	}
}

// Demo main update
void UpdateDemo()
{
	static DWORD lastTimeAnimationChanged=timeGetTime();
	static DWORD lastUpdateTime=timeGetTime();
	
	DWORD timeElapsedSinceAnimationChange=timeGetTime()-lastTimeAnimationChanged;
	
	// Periodically advance to the next animation set
	if (timeElapsedSinceAnimationChange>kTimeBeforeNextAnimation)
	{
		gCurrentAnimationSet++;
		if (gCurrentAnimationSet>=gNumAnimationSets)
			gCurrentAnimationSet=0;

		gXAnimator->ChangeAnimationSet(gModelId,gCurrentAnimationSet,kTransitionTime);
		lastTimeAnimationChanged=timeGetTime();
	}

	gD3dDevice->Clear(0,NULL,D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,50),1.0f,0);
	gD3dDevice->BeginScene();
	

	// Render our animated model
	D3DXMATRIX world;
	D3DXMatrixIdentity(&world);

	// Slowly rotate it (comment out to keep it still)
	D3DXMatrixRotationY(&world, timeGetTime()%3600*0.00174533f);

	DWORD timeElapsedSinceLastUpdate=timeGetTime()-lastUpdateTime;

	gXAnimator->Render(gModelId,world,(float)timeElapsedSinceLastUpdate);

	lastUpdateTime=timeGetTime();

	gD3dDevice->EndScene();
	gD3dDevice->Present( NULL, NULL, NULL, NULL ); 
}

// Load the model using XAnimator and grab data
bool LoadModel(const std::string &filename)
{
	if (!gXAnimator->LoadXFile(filename,&gModelId))
		return false;

	gNumAnimationSets=gXAnimator->GetNumberOfAnimationSets(gModelId);
	gCurrentAnimationSet=1;

	// First animation set to start with - note: v.small transition time as it is the first
	gXAnimator->ChangeAnimationSet(gModelId,gCurrentAnimationSet,0.1f);

	return true;
}